/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <stdio.h>
#include <errno.h>
#include <sys/mman.h>

#include <mach/inline.h>
#include <mach/vm_param.h>
#include <mach/machine/vm_types.h>
#include <mach/machine/far_ptr.h>
#include <mach/machine/proc_reg.h>
#include <mach/machine/eflags.h>
#include <mach/machine/code16.h>
#include <mach/machine/asm.h>
#include <mach/machine/dpmi.h>

#include "real.h"
#include "prog.h"
#include "misc.h"
#include "ldt.h"
#include "debug.h"
#include "vm_param.h"
#include "signal.h"
#include "irq.h"
#include "config.h"
#include "i16_crt0.h"

#ifdef ENABLE_DPMI

#define dpmi_err(exp) \
	({ dpmi_error_t err = (exp); if (err) die("DPMI error %04x", err); })
#define i16_dpmi_err(exp) \
	({ dpmi_error_t err = (exp); if (err) i16_die("DPMI error"); })

static char bounce_code[1+4+2] = {0xea};

static unsigned short kernel_cs, kernel_ds;

unsigned short dpmi_prog_cs, dpmi_prog_ds;
struct far_pointer_32 dpmi_kernel_stack;
struct far_pointer_32 dpmi_prog_stack;

static unsigned dpmi_page_size;

#define SET_EXCHANDLER(n)					\
{								\
	extern void dpmi_exchandler_##n(void);			\
	struct far_pointer_32 vec = {				\
		(unsigned)dpmi_exchandler_##n, kernel_cs};	\
	dpmi_set_exception_handler(n, &vec);			\
}

static vm_offset_t prog_limit;
static vm_offset_t prog_brk;
static vm_offset_t prog_pbrk;

/* True if the program wants all allocated memory to be wired.  */
static int mlock_future;

static dpmi_error_t alloc(
	vm_offset_t	desired_pva,
	vm_size_t	size,
	int		locked,
	vm_offset_t	*out_pva,
	unsigned	*out_handle)
{
	dpmi_error_t err;
	vm_offset_t desired_la, la, pva;
	unsigned handle;

	/* First try to allocate at the exact location specified.  */
	if (desired_pva == 0) desired_pva = prog_pbrk;
	desired_la = progva_to_la(desired_pva);
	err = dpmi_allocate_linear_memory(desired_la, size, 0, &la, &handle);
	if (!err)
	{
		assert(la == desired_la);
		assert(la > progva_to_la(0));
	}
	else
	{
		retry:

		/* Allocate linear memory request failed -
		   try using the old allocation method.  */
		err = dpmi_allocate_memory(size, &la, &handle);
		if (err) return err;

		/* If a block is returned that is lower than the block
		   allocated for the program itself,
		   then just leave the block allocated and try again.
		   This is horrible I know, but what else can be done...
		   Windows Sucks.  */
		if (la < progva_to_la(0))
		{
			goto retry;
		}
	}
	pva = la_to_progva(la);

	/* Wire down the memory if so requested.  */
	if (locked)
	{
		err = dpmi_lock_linear_region(la, size);
		if (err) return err;
	}

	/* Adjust the program's data segment limit if necessary.  */
	if (pva + size > prog_limit)
	{
		err = dpmi_set_segment_limit(dpmi_prog_ds, pva + size - 1);
		if (err) return err;
		prog_limit = pva + size;
	}

	*out_pva = pva;
	*out_handle = handle;
	return 0;
}

void dpmi_prog_run()
{
	extern void dpmi_syscall_entry();

	unsigned size;
	unsigned prog_linear, stack_pva;
	unsigned prog_handle, stack_handle;
	vm_offset_t init_esp;
	struct i386_descriptor seg_desc;
	dpmi_error_t dpmi_err;
	int i;

	/* Find the basic in-memory program size.  */
	prog_limit = size = 0;
	for (i = 0; i < MAX_PROG_SEGS; i++)
	{
		vm_offset_t seg_end;
		if (prog_seg[i].mem_size == 0)
			continue;
		seg_end = prog_seg[i].mem_start + prog_seg[i].mem_size;
		if (seg_end > size)
			size = seg_end;
		if ((seg_end > prog_limit)
		    && (prog_seg[i].mem_prot & VM_PROT_EXECUTE))
		{
			prog_limit = seg_end;
		}
	}
	size = round_page(size);

	/* Allocate a linear memory region for the program itself.  */
	dpmi_err(dpmi_allocate_memory(size, &prog_linear, &prog_handle));

	/* Set up prog_va and the break pointers appropriately.  */
	prog_va = phys_mem_va + prog_linear;
	prog_brk = prog_pbrk = size;

	/* Create code and data segment selectors for the program.  */
	dpmi_err(dpmi_allocate_descriptors(1, &dpmi_prog_cs));
	dpmi_err(dpmi_get_descriptor(kernel_cs, &seg_desc));
	fill_descriptor_base(&seg_desc, prog_linear);
	fill_descriptor_limit(&seg_desc, prog_limit - 1);
	dpmi_err(dpmi_set_descriptor(dpmi_prog_cs, &seg_desc));
	dpmi_err(dpmi_create_code_segment_alias(dpmi_prog_cs, &dpmi_prog_ds));

	/* Allocate a region for the program's stack.
	   Try to leave a one-page red zone to catch stack overflows.
	   The alloc() routine will update the segment descriptor limits.  */
	dpmi_err(alloc(size + dpmi_page_size, PROG_STACK_SIZE, 0,
			&stack_pva, &stack_handle));
	init_esp = stack_pva + PROG_STACK_SIZE;

	/* Set the initial break pointers one page beyond the stack,
	   leaving a second red zone between stack and heap.  */
	prog_pbrk = prog_brk = init_esp + dpmi_page_size;

	/* Selector 7 needs to look like a call gate to the program,
	   but DPMI doesn't allow us to create real call gates.
	   So instead set it up as a real code segment descriptor
	   with a base address such that offset 0
	   points to another ljmp instruction to bounce to the real cs.  */
	*((unsigned*)(bounce_code+1)) = (unsigned)dpmi_syscall_entry;
	*((unsigned short*)(bounce_code+1+4)) = kernel_cs;
	fill_descriptor_base(&seg_desc, (unsigned)bounce_code - phys_mem_va);
	fill_descriptor_limit(&seg_desc, sizeof(bounce_code)-1);
	dpmi_err(dpmi_set_descriptor(USER_SCALL, &seg_desc));

	/* DPMI doesn't support demand loading,
	   so read the entire program all in one shot.  */
	for (i = 0; i < MAX_PROG_SEGS; i++)
	{
		if (prog_seg[i].file_size > 0)
		{
			prog_read(prog_seg[i].file_start,
				  progva_to_va((void*)prog_seg[i].mem_start),
				  prog_seg[i].file_size);
		}
		if (prog_seg[i].mem_size > prog_seg[i].file_size)
		{
			bzero(progva_to_va((void*)(prog_seg[i].mem_start
						   + prog_seg[i].file_size)),
			      prog_seg[i].mem_size - prog_seg[i].file_size);
		}
	}

	/* The stack grows down from the initial break pointer.  */
	dpmi_prog_stack.seg = dpmi_prog_ds;
	dpmi_prog_stack.ofs = prog_push_args_env(init_esp);

	asm volatile("
		lss	("SEXT(dpmi_prog_stack)"),%%esp
		pushl	%0
		pushl	%1
		movw	%2,%%ds
		movw	%2,%%es
		lret
	" :
	  : "r" (dpmi_prog_cs), "r" (prog_info.entry), "r" (dpmi_prog_ds));

	while(1);
}

static void dpmi_real_int(int intnum, struct real_call_data *rcd)
{
	dpmi_err(dpmi_simulate_real_mode_interrupt(intnum, rcd));
}

static void dpmi_exit(int rc)
{
	asm volatile("
		int	$0x21
	" : : "a" (0x4c00 | (rc & 0xff)));
}

/* This implementation of sbrk isn't quite up to Unix sbrk standards,
   in that in that it doesn't guarantee that the next returned memory block
   will be immediately after the previous one.
   DPMI doesn't provide an easy way of implementing a real Unix sbrk, unfortunately.
   (We could sort of do this by growing the program's initial memory segment,
   except that resizing a memory region can cause it to move,
   and that would change the relative location of physical memory as the program sees it,
   which would screw things up really badly for many programs.)

   Many Unixish malloc packages depend on this property, including GNU malloc I believe.
   Fortunately, the BSD malloc package only assumes Unix sbrk semantics
   for the purpose of implementing its nice natural alignment properties.
   it'll still work fine under this slightly broken sbrk;
   the memory blocks it returns just won't be quite so nicely aligned.  */
/* XXX currently panics on failure */
static int dpmi_prog_sbrk(int mem_needed, vm_offset_t *out_prog_va)
{
	/* Note: better not call this variable prog_va;
	   that clashes with the outer-scope prog_va used by la_to_progva.  */
	vm_offset_t prog_ret;

	assert(prog_pbrk >= prog_brk);
	if ((prog_brk + mem_needed) <= prog_pbrk)
	{
		/* This allocation fits onto the remainder of the last page.  */
		prog_ret = prog_brk;
		prog_brk += mem_needed;
	}
	else
	{
		dpmi_error_t err;
		vm_size_t mem_needed_p;
		vm_offset_t prog_new_mem;
		unsigned handle;

		/* Allocate some new pages of memory.  */
		mem_needed_p = round_page(mem_needed);
		err = alloc(prog_pbrk, mem_needed_p, mlock_future,
			    &prog_new_mem, &handle);
		if (err) return err;

		/*printf("allocated %08x at prog %08x (lin %08x)\n",
			mem_needed_p, prog_new_mem, lin_new_mem);*/
		if (prog_new_mem == prog_pbrk)
		{
			/* We were given memory extending the previous hunk.
			   Good - we can act exactly like Unix sbrk.  */
			prog_ret = prog_brk;
			prog_pbrk += mem_needed_p;
			prog_brk += mem_needed;
		}
		else
		{
			/* We were given memory somewhere else.
			   We'll just have to return that.  */
			prog_ret = prog_new_mem;
			prog_pbrk = prog_new_mem + mem_needed_p;
			prog_brk = prog_new_mem + mem_needed;
		}
	}
	assert(prog_pbrk >= prog_brk);

	/* DPMI memory isn't guaranteed to be cleared,
	   and some Unix software (e.g. the BSD malloc)
	   expect sbrk-allocated memory to be zero-filled,
	   so we have to clear the memory ourselves.  */
	memset((void*)progva_to_va(prog_ret), 0, mem_needed);

	*out_prog_va = prog_ret;
	return 0;
}

static int dpmi_prog_mmap_phys(vm_offset_t pa, vm_size_t size,
				vm_offset_t *inout_pva)
{
	dpmi_error_t err;
	vm_offset_t page_pa;
	unsigned count;
	vm_offset_t page_pva;
	unsigned handle;

	page_pa = pa & ~(dpmi_page_size - 1);
	size += pa - page_pa;
	count = (size + dpmi_page_size - 1) / dpmi_page_size;
	size = count * dpmi_page_size;

	if (prog_limit == 0xffffffff)
		goto do_direct;

	/* Allocate a sufficiently large region of memory.  */
	err = alloc(*inout_pva, size, 0/*XXX?*/, &page_pva, &handle);
	if (err) return err;

	/* Map the requested physical memory over that memory block.  */
	err = dpmi_map_conventional_memory(handle, 0, page_pa, count);
	if (err)
	{
		/* That didn't work,
		   probably just because this DPMI host doesn't support it.
		   Try one last kludgy solution:
		   extend the program's data segment limit
		   so the data segment wraps around to low linear memory,
		   giving it access to conventional memory that way.
		   This works under Windows 3.1, for example,
		   but not under systems like OS/2 2.0
		   which chop off the limit below the kernel boundary.  */

		struct i386_descriptor desc;

		/* First of all, free the memory block we grabbed,
		   since we certainly won't need it now regardless.  */
		dpmi_free_memory(handle);

		/* Try setting the limit.  */
		if (dpmi_set_segment_limit(dpmi_prog_ds, 0xffffffff))
			return err;

		/* Make sure it hasn't been silently chopped.
		   There's no get_limit call,
		   so we have to get the whole descriptor.  */
		if (dpmi_get_descriptor(dpmi_prog_ds, &desc))
			return err;
		if (!(desc.granularity & SZ_G)
		    || (desc.limit_low != 0xffff)
		    || (desc.limit_high != 0xf))
			return err;

		/* Yippee - it worked.
		   Now we won't have to go through this again.  */
		prog_limit = 0xffffffff;

		do_direct:
		page_pva = va_to_progva(phystokv(page_pa));
	}

	*inout_pva = page_pva + (pa - page_pa);
	return 0;
}

static int dpmi_prog_mlock(vm_offset_t pva, vm_size_t len)
{
	/* Note that this isn't quite a proper POSIX implementation of mlock/munlock,
	   because DPMI specifies a lock count kept for each page, whereas POSIX doesn't.
	   Big fat hairy deal.  */
	dpmi_err(dpmi_lock_linear_region(progva_to_la(pva), len));
}

static int dpmi_prog_munlock(vm_offset_t pva, vm_size_t len)
{
	dpmi_err(dpmi_unlock_linear_region(progva_to_la(pva), len));
}

static int dpmi_prog_mlockall(int flags)
{
	if (flags & MCL_FUTURE)
		mlock_future = 1;

	if (flags & MCL_CURRENT)
		panic("dpmi_prog_mlockall");

	return 0;
}

static int dpmi_prog_munlockall()
{
	panic("dpmi_prog_munlock");
}

static struct far_pointer_32 dpmi_old_handlers[16];

extern vm_offset_t dpmi_sig_int_handlers[16];
extern vm_offset_t dpmi_sig_int_post_handlers[16];

static int dpmi_prog_irq_catch(int irq, struct signode *n)
{
	/* It appears DPMI is just plain too stupid
	   to support dispatching interrupts as signals,
	   because it doesn't define anything about the interrupt stack frame
	   that allows us to redirect the flow of the program
	   if it's currently executing in user mode.  */
	return ENOSYS;
#if 0
	struct far_pointer_32 vec = {
		dpmi_sig_int_handlers[irq], kernel_cs};
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);
	int err;

	assert(irq >= 0); assert(irq < 16);
	assert(!(get_eflags() & EFL_IF));

	/* Save the old protected-mode interrupt handler.  */
	if (err = dpmi_get_interrupt_handler(intvec, &dpmi_old_handlers[irq]))
		return err;

	/* Set the protected-mode interrupt handler.  */
	if (err = dpmi_set_interrupt_handler(intvec, &vec))
		return err;

	/* This routine gets called after the user signal hander returns;
	   it reenables the hardware interrupt.  */
	n->post_handler = (void*)dpmi_sig_int_post_handlers[irq];

	return 0;
#endif
}

extern vm_offset_t dpmi_fastint_handlers[16];
extern void dpmi_fastint_return();

struct
{
	vm_offset_t eip;
	vm_offset_t esp;
} dpmi_fastint_nodes[16];

static int dpmi_prog_irq_catch_fast(int irq, vm_offset_t eip, vm_offset_t esp)
{
	struct far_pointer_32 vec = {
		dpmi_fastint_handlers[irq], kernel_cs};
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);
	struct {
		unsigned handler_eip;	/* set on each interrupt */
		unsigned handler_cs;	/* set on each interrupt */
		unsigned ret_addr;	/* short return addr to bounce code */
		unsigned saved_esp;	/* set on each interrupt */
		unsigned saved_ss;	/* set on each interrupt */
	} *fastint_stack;
	struct far_pointer_32 *bounce_code;
	int err;

	assert(irq >= 0); assert(irq < 16);
	assert(!(get_eflags() & EFL_IF));

	/* Stash the handler eip and esp.
	   The fastint glue code in locore.S expects esp
	   to be adjusted by 16 bytes for the initial stack frame.  */
	esp -= sizeof(*fastint_stack);
	dpmi_fastint_nodes[irq].eip = eip;
	dpmi_fastint_nodes[irq].esp = esp;

	/* Set up the permanent stuff on the stack.  */
	fastint_stack = (void*)progva_to_va(esp);
	fastint_stack->ret_addr = prog_info.entry - 32;

	/* Fill in the magic bounce code in the program's crt0.  */
	assert(*((unsigned char*)progva_to_va(prog_info.entry - 32)) == 0xea);
	bounce_code = (struct far_pointer_32*)
		progva_to_va(prog_info.entry - 32 + 1);
	bounce_code->ofs = (vm_offset_t)dpmi_fastint_return;
	bounce_code->seg = kernel_cs;

	/* Save the old protected-mode interrupt handler.  */
	if (err = dpmi_get_interrupt_handler(intvec, &dpmi_old_handlers[irq]))
		return err;

	/* Set the protected-mode interrupt handler.  */
	if (err = dpmi_set_interrupt_handler(intvec, &vec))
		return err;

	return 0;
}

static void dpmi_prog_irq_free(int irq)
{
	int intvec = (irq < 8) ? (irq_master_base + irq)
			       : (irq_slave_base + irq - 8);

	if (dpmi_old_handlers[irq].seg)
		dpmi_set_interrupt_handler(intvec, &dpmi_old_handlers[irq]);
}

static const struct progvec dpmi_progvec =
{
	dpmi_prog_run,
	dpmi_prog_sbrk,
	dpmi_prog_mmap_phys,
	dpmi_prog_mlock,
	dpmi_prog_munlock,
	dpmi_prog_mlockall,
	dpmi_prog_munlockall,
	dpmi_prog_irq_catch,
	dpmi_prog_irq_free,
	dpmi_prog_irq_catch_fast,
	dpmi_prog_irq_free,
};

CODE16

#include <mach/machine/proc_reg.h>

static struct far_pointer_16 pmode_entry;

void i16_dpmi_check()
{
	extern vm_offset_t dos_mem_phys_free_mem;
	extern vm_size_t dos_mem_phys_free_size;
	extern unsigned short environ_seg, psp_seg;

	unsigned short rc, dpmi_flags, host_data_paras, dpmi_version;
	unsigned char proc_type;
	unsigned short host_data_seg;
	struct i386_descriptor seg_desc;

	/* Check for the presence of DPMI.  */
	asm volatile("
		pushw	%%es
		int	$0x2f
		movw	%%es,"SEXT(pmode_entry)"+2
		movw	%%di,"SEXT(pmode_entry)"
		popw	%%es
	"
		: "=a" (rc),
		  "=b" (dpmi_flags),
		  "=c" (proc_type),
		  "=d" (dpmi_version),
		  "=S" (host_data_paras)
		: "a" ((unsigned short)0x1687)
		: "edi");
	if (rc != 0)
		return;
	if (!(dpmi_flags & 1))
		i16_die("DPMI host doesn't support 32-bit programs!");

	/* We shouldn't have gotten this far if we're not on at least a 386.  */
	i16_assert(proc_type >= 3);

	do_debug(i16_puts("DPMI detected"));

	/* Allocate the DPMI host data area.  */
	if ((!dos_mem_phys_free_mem)
	    || (dos_mem_phys_free_size < host_data_paras * 16))
		i16_die("Not enough DOS memory for DPMI host data area");
	i16_assert((dos_mem_phys_free_mem & 15) == 0);
	host_data_seg = dos_mem_phys_free_mem / 16;
	dos_mem_phys_free_mem += host_data_paras * 16;
	dos_mem_phys_free_size -= host_data_paras * 16;

	/* Switch to protected mode.  */
	i16_dpmi_err(dpmi_switch_to_pmode(&pmode_entry, host_data_seg));

	/* Allocate the syscall gate descriptor before anything else,
	   just to be sure it doesn't become unavailable.  */
	i16_dpmi_err(dpmi_allocate_specific_descriptor(USER_SCALL));

	/* Create the kernel code and data segment selectors.  */
	i16_dpmi_err(dpmi_allocate_descriptors(1, &kernel_cs));
	i16_dpmi_err(dpmi_get_descriptor(get_cs(), &seg_desc));
	seg_desc.granularity |= SZ_32;
	i16_dpmi_err(dpmi_set_descriptor(kernel_cs, &seg_desc));
	i16_dpmi_err(dpmi_set_segment_limit(kernel_cs, 0xffffffff));
	i16_dpmi_err(dpmi_create_code_segment_alias(kernel_cs, &kernel_ds));

	/* Find the appropriate flags for the mode in which we're running.  */
	real_call_data.flags = get_eflags();

	/* Set the appropriate function pointers for DPMI mode.  */
	real_int = dpmi_real_int;
	real_exit = dpmi_exit;
	progvec = &dpmi_progvec;

	/* Switch to 32-bit mode.  */
	asm volatile("
		pushl	%0
		pushl	$1f
		lret
		.code32
	1:
		movw	%1,%%ds
		movw	%1,%%es
		movw	%1,%%ss
	" : : "r" ((unsigned long)kernel_cs), "r" ((unsigned short)kernel_ds));

	dpmi_kernel_stack.seg = kernel_ds;
	dpmi_kernel_stack.ofs = (vm_offset_t)crt0_stack + CRT0_STACK_SIZE;

	dpmi_err(dpmi_get_page_size(&dpmi_page_size));

	/* Install our own set of exception handlers
	   so we can get a nice register/stack dump if something goes wrong.  */
	SET_EXCHANDLER(0);
	SET_EXCHANDLER(1);
	SET_EXCHANDLER(2);
	SET_EXCHANDLER(3);
	SET_EXCHANDLER(4);
	SET_EXCHANDLER(5);
	SET_EXCHANDLER(6);
	SET_EXCHANDLER(7);
	SET_EXCHANDLER(8);
	SET_EXCHANDLER(9);
	SET_EXCHANDLER(10);
	SET_EXCHANDLER(11);
	SET_EXCHANDLER(12);
	SET_EXCHANDLER(13);
	SET_EXCHANDLER(14);
	SET_EXCHANDLER(15);

#if 0 /* XXX kludge to get around DOSEMU limitation */
{
	extern unsigned short environ_seg, psp_seg;
	static char zeros[256];
	psp_seg = environ_seg = (vm_offset_t)kvtophys(zeros)/16+1;
}
#endif

	/* Find the PICs, so we can set and restore IRQ handlers.  */
	{
		struct dpmi_version_status vs;
		dpmi_get_version(&vs);
		/*
		printf("DPMI %d.%d flags %04x proc %d master %02x slave %02x\n",
			vs.major_version, vs.minor_version,
			vs.flags, vs.processor_type,
			vs.master_pic_base, vs.slave_pic_base);
		*/
		irq_master_base = vs.master_pic_base;
		irq_slave_base = vs.slave_pic_base;
	}

	/* Now that we can access all physical memory,
	   collect the memory regions we discovered while in 16-bit mode
	   and add them to our free memory list.
	   We can't do this before now because the free list nodes
	   are stored in the free memory itself,
	   which is probably out of reach of our 16-bit segments.  */
	phys_mem_collect();

	/* Jump into the mode-independent code to load and run the program.  */
	prog_start();

	CODE16
}

void i16_dpmi_shutdown()
{
}

#endif ENABLE_DPMI
