/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1995 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */
/* Embedded File System:
   serves up read-only files embedded in our DOS executable.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "efs.h"
#include "prog.h"
#include "dos_io.h"
#include "vm_param.h"

/* Table of embedded file names, offsets, and sizes.  */
struct digentry
{
	const char *name;
	off_t offset;
	size_t size;
};
static struct digentry *digentry;
static int digentries;

/* DOS file descriptor for our own executable.  */
static dos_fd_t exe_fd;

/* Current offset within that file,
   to avoid unnecessary seeking.  */
static off_t exe_off;

void efs_init()
{
	extern char **argv;
	unsigned short exehdr[3];
	vm_size_t actual, newpos;
	off_t off;
	char hdr[32];
	char *digest;
	int diglines, digsize;
	int i;

	/* Open our executable file.  */
	if (dos_open(argv[0], O_RDONLY, 0, &exe_fd))
		return;

	/* Read the first part of the EXE header,
	   to find out how long the original EXE file is.  */
	if (dos_read(exe_fd, exehdr, sizeof(exehdr), &actual)
	    || (actual != sizeof(exehdr))
	    || (exehdr[0] != 0x5a4d))
		return;
	off = (off_t)exehdr[2] * 512;
	/*printf("size of EXE file: %d blocks (%d bytes)\n", exehdr[2], off);*/

	/* Read the digest header, if there is one.
	   It will tell us the size of the digest,
	   and the number of lines (entries) it contains.  */
	if (dos_seek(exe_fd, off, SEEK_SET, &newpos)
	    || (newpos != off)
	    || (dos_read(exe_fd, hdr, sizeof(hdr), &actual))
	    || (actual != sizeof(hdr))
	    || (sscanf(hdr, " %d %*d %d", &diglines, &digsize) != 2)
	    || (digsize <= 0) || (diglines <= 0))
		goto corrupt;
	/*printf("digest size: %d lines: %d\n", digsize, diglines);*/
	off += 512;

	/* Read the digest itself.  */
	digest = malloc(digsize);
	if (digest == 0)
		panic("efs_init: out of memory");
	if (dos_seek(exe_fd, off, SEEK_SET, &newpos)
	    || (newpos != off)
	    || dos_read(exe_fd, digest, digsize, &actual)
	    || (actual != digsize))
		goto corrupt;
	off += (digsize + 511) & ~511;

	/* Find the filename of the program to load and run.  */
	prog_filename = digest;
	if (!(digest = strchr(digest, '\n')))
		goto corrupt;
	*digest = 0;
	digest++; diglines--;
	/*printf("prog_filename: `%s'\n", prog_filename);*/

	/* Allocate and initialize the digentry array.  */
	digentries = diglines;
	digentry = malloc(sizeof(struct digentry) * digentries);
	if (digentry == 0)
		panic("efs_init: out of memory");
	for (i = 0; i < digentries; i++)
	{
		digentry[i].offset = off;
		digentry[i].size = strtoul(digest, &digest, 10);
		if (*digest != ' ')
			goto corrupt;
		digest++;
		digentry[i].name = digest;
		if (!(digest = strchr(digest, '\n')))
			goto corrupt;
		*digest = 0;
		digest++;
		off += (digentry[i].size + 511) & ~511;
	}

	return;

	corrupt:
	/*printf("Warning: EFS digest is corrupt\n");*/
	digentry = 0; digentries = 0;
	dos_close(exe_fd);
	return;
}

int efs_open(const char *name, int flags, mode_t mode, struct fd_efs *fd)
{
	unsigned i;

/*printf("efs_open %s %d\n", name, flags);*/
	/* All files on this "file system" are read-only.  */
	if ((flags & O_ACCMODE) != O_RDONLY)
		return EROFS;

	/* Search for the specified filename.  */
	for (i = 0; ; i++)
	{
		if (i >= digentries)
		{
#if 0
printf(" %s not found; entries (%d):\n", name, digentries);
for (i = 0; i < digentries; i++)
	printf("  %d %s %d\n", i, digentry[i].name, digentry[i].size);
#endif
			return ENOENT;
		}
		if (!strcmp(name, digentry[i].name))
			break;
	}

	/* Found it - just record the index number.  */
	fd->idx = i;
	fd->ofs = 0;

	return 0;
}

int efs_close(struct fd_efs *fd)
{
	/* Nothing to do */
	return 0;
}

int efs_read(struct fd_efs *fd, void *buf, size_t size, size_t *actual)
{
	struct digentry *ent = &digentry[fd->idx];
	off_t new_exe_off, act_new_exe_off;
	int err;

/*printf("efs_read %s, %d\n", ent->name, size);*/
	/* If we're at or beyond the end of the file, return EOF.  */
	if (fd->ofs >= ent->size)
	{
		*actual = 0;
		return 0;
	}

	/* Chop the actual size according to the size of the embedded file.  */
	if (size > ent->size - fd->ofs)
		size = ent->size - fd->ofs;

	/* Seek to the proper position in the EXE file, if necessary.  */
	new_exe_off = ent->offset + fd->ofs;
/*printf(" exe_off %d new_exe_off %d\n", exe_off, new_exe_off);*/
	if (exe_off != new_exe_off)
	{
		if (err = dos_seek(exe_fd, new_exe_off, SEEK_SET, &exe_off))
		{
			exe_off = 0;
			return err;
		}
		if (exe_off != new_exe_off)
		{
			return EIO;
		}
	}

	/* Read the data.  */
	/* XXX potential seek pointer race condition with pagein? */
	if (err = dos_read(exe_fd, buf, size, actual))
	{
		exe_off = 0;
		return err;
	}
/*printf(" dos_read %d -> %d\n", size, *actual);*/
	exe_off += *actual;
	if (*actual != size)
	{
		return EIO;
	}

	fd->ofs += size;
	return 0;
}

int efs_seek(struct fd_efs *fd, off_t offset, int whence, off_t *newpos)
{
/*printf("efs_seek %s, %d,%d\n", digentry[fd->idx].name, offset, whence);*/
	switch (whence)
	{
		case SEEK_SET:
			fd->ofs = offset;
			break;
		case SEEK_CUR:
			fd->ofs += offset;
			break;
		case SEEK_END:
			fd->ofs = digentry[fd->idx].size + offset;
			break;
		default:
			return EINVAL;
	}
/*printf(" newpos %d\n", fd->ofs);*/
	*newpos = fd->ofs;
	return 0;
}

int efs_fstat(struct fd_efs *fd, struct stat *st)
{
	memset(st, 0, sizeof(*st));
	st->st_nlink = 1;
	st->st_mode = S_IRWXU | S_IRWXG | S_IRWXO | S_IFREG;
	st->st_size = digentry[fd->idx].size;
	st->st_blocks = (st->st_size + 511) / 512;
	st->st_blksize = 512;
	return 0;
}

