/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <mach/machine/vm_types.h>
#include <mach/machine/proc_reg.h>
#include <mach/machine/paging.h>
#include <mach/lmm.h>
#include <malloc.h>
#include <errno.h>

#include "config.h"
#include "vm_param.h"
#include "phys_mem.h"
#include "debug.h"

#define KERNEL_PTE (INTEL_PTE_VALID | INTEL_PTE_USER | INTEL_PTE_WRITE)

static vm_offset_t pdir;

/*
 *	Convert page table entry to kernel virtual address
 */
#define ptetokv(a)	(phystokv(pte_to_pa(a)))

/*
 *	Allocate a page, paging something out if necessary.
 */
vm_offset_t page_alloc()
{
	vm_offset_t page_va;

	page_va = (vm_offset_t)lmm_alloc_aligned(&malloc_lmm, PAGE_SIZE, 0, PAGE_SHIFT, 0);
	if (!page_va)
		return 0;

	assert((page_va & PAGE_MASK) == 0);

	return kvtophys(page_va);
}

/*
 *	Routine:	pmap_page_table_page_alloc
 *
 *	Allocates a new physical page to be used as a page-table page.
 *
 *	Must be called with the pmap system and the pmap unlocked,
 *	since these must be unlocked to use vm_page_grab.
 */
vm_offset_t page_table_page_alloc()
{
	vm_offset_t pa;

	/*
	 *	Allocate a VM page for the level 2 page table entries.
	 *	XXX
	 */
	pa = page_alloc();
	if (!pa)
		return 0;

	/*
	 *	Zero the page.
	 */
	bzero(phystokv(pa), PAGE_SIZE);

	return pa;
}

/*
 *	Insert the given physical page (p) at
 *	the specified virtual address (v) in the
 *	target physical map with the protection requested.
 *
 *	If specified, the page will be wired down, meaning
 *	that the related pte can not be reclaimed.
 *
 *	NB:  This is the only routine that MAY NOT lazy-evaluate
 *	or lose information.  That is, this routine must actually
 *	insert this page into the given map NOW.
 */
int pmap_map_page(vm_offset_t la, pt_entry_t mapping)
{
	pt_entry_t *pde, *pte;

	/* Find the page directory entry.  */
	pde = &((pt_entry_t*)phystokv(pdir))[lin2pdenum(la)];
	assert(pde != 0);

	/* Find the page table entry,
	   creating the page table if necessary.  */
	if (!(*pde & INTEL_PTE_VALID))
	{
		vm_offset_t ptable;

		/* Allocate a new page table.  */
		ptable = page_table_page_alloc();
		if (!ptable)
			return ENOMEM;

		/* Set the page directory entry to point to it.  */
		*pde = pa_to_pte(ptable)
			| INTEL_PTE_VALID | INTEL_PTE_USER | INTEL_PTE_WRITE;
	}
	assert(*pde & INTEL_PTE_VALID);
	assert(pte_to_pa(*pde));
	assert(ptetokv(*pde));
	pte = &((pt_entry_t*)ptetokv(*pde))[lin2ptenum(la)];

	/* Insert the mapping.
	   If there was already a valid mapping,
	   leave the accessed and dirty bits unmodified.  */
	if ((mapping & INTEL_PTE_VALID) && (*pte & INTEL_PTE_VALID))
		mapping |= *pte & (INTEL_PTE_MOD | INTEL_PTE_REF);
	*pte = mapping;

	return 0;
}

pt_entry_t pmap_get_pte(vm_offset_t la)
{
	pt_entry_t *pde, *pte;

	/* Find the page directory entry.  */
	pde = &((pt_entry_t*)phystokv(pdir))[lin2pdenum(la)];
	assert(pde != 0);

	/* Find the page table entry.  */
	if (!(*pde & INTEL_PTE_VALID))
		return 0;
	assert(*pde & INTEL_PTE_VALID);
	assert(pte_to_pa(*pde));
	assert(ptetokv(*pde));

	pte = &((pt_entry_t*)ptetokv(*pde))[lin2ptenum(la)];

	return *pte;
}

void raw_paging_enable()
{
	i386_paging_enable(pdir);
}

void raw_paging_disable()
{
	i386_paging_disable();
}

void raw_paging_init()
{
	vm_offset_t addr;

	pdir = page_table_page_alloc();
	if (pdir == 0)
		panic("not enough memory");

	/*
	 * Map all known physical memory at linear address 0.
	 */
	for (addr = 0; addr < phys_mem_max; addr += PAGE_SIZE)
	{
		if (pmap_map_page(addr, addr | KERNEL_PTE))
			panic("not enough memory");
	}
}

#ifdef ENABLE_VCPI

void vcpi_paging_init(vm_offset_t vcpi_pdir, vm_offset_t phys_high_start)
{
	vm_offset_t addr;

	/* The VCPI initilization code creates the first page directory,
	   and we must use it forever.  */
	pdir = vcpi_pdir;

	assert(!(phys_high_start & PAGE_MASK));

	for (addr = phys_high_start; addr < phys_mem_max; addr += PAGE_SIZE)
	{
		if (pmap_map_page(addr, addr | KERNEL_PTE))
			panic("not enough memory");
	}
}

#endif ENABLE_VCPI

