/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1995 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <mach/proc_ops.h>
#include <mach/machine/eflags.h>
#include <mach/machine/pio.h>
#include <errno.h>
#include <sys/signal.h>

#include "prog.h"
#include "trap.h"
#include "signal.h"
#include "remote-gdb.h"
#include "debug.h"

/* Flag: nonzero if there may be new signals pending that need to be handled.  */
int new_sigs;

static struct sigaction sigact[NSIG];

struct signode sig_pending_list;
sigset_t sig_mask;

/* Signal nodes for generic unqueued signals */
static struct signode sig_gen_nodes[NSIG];

/* Signal nodes for queued IRQ-generated signals */
struct signode sig_irq_nodes[16];
static int irq_orig_enable_mask;

void sig_init(void)
{
	sig_pending_list.next = sig_pending_list.prev = &sig_pending_list;
}

static void sig_clear(struct signode *cn)
{
	struct signode *n;

	for (n = sig_pending_list.next; n != &sig_pending_list; n = n->next)
	{
		assert(n);
		if (n == cn)
		{
			sig_remove(n);
			n->post_handler(n);
			return;
		}
	}

	/* XXX but what if the signal is currently being handled in user space? */
}

void check_sigs(int dummy)
{
	struct trap_state *ts = (struct trap_state*)&dummy;
	struct signode *n, *nn;

	restart:
	new_sigs = 0;
	for (n = sig_pending_list.next; n != &sig_pending_list; n = nn)
	{
		int signo = n->signo;

		assert(n);
		assert(signo > 0); assert(signo < NSIG);
		nn = n->next;

		/* See if the signal is unblocked.  */
		if (!(sig_mask & n->sigmask))
		{
			struct sigaction *act = &sigact[signo];

			/* Remove the node from the pending list.  */
			sig_remove(n);

			/* Handle it appropriately.  */
			if (act->sa_handler == SIG_DFL)
			{
				/* Give the debugger a whack at it.
				   This really should be done above
				   regardless of the signal's action;
				   however, this really bogs things down
				   when signals are occurring rapidly.  */
				if (rgdb_signal(&signo, ts))
				{
					if (signo == 0)
					{
						/* Consume this signal
						   and continue to the next.  */
						n->post_handler(n);
						continue;
					}

					/* We might have been instructed
					   to continue with a different signal,
					   so adjust appropriately.  */
					act = &sigact[signo];
				}

				printf("Unexpected signal %d\n", signo);
				trap_dump_die(ts);
			}
			else if (act->sa_handler == SIG_IGN)
			{
				/* Dispose of the signode immediately.  */
				n->post_handler(n);
			}
			else
			{
				struct sigcontext *sc;

				/* Push a sigcontext frame onto the user stack,
				   and modify the trap_state
				   to jump to the signal handler.  */
				ts->esp -= sizeof(*sc);
				sc = (struct sigcontext*)progva_to_va(ts->esp);
				sc->sc_return_addr = prog_info.entry - 16;
				sc->sc_signo = signo;
				sc->sc_value = n->value;
				sc->sc_mask = sig_mask;
				sc->sc_eip = ts->eip;
				sc->sc_efl = ts->eflags;
				sc->sc_edx = ts->edx;
				sc->sc_ecx = ts->ecx;
				sc->sc_eax = ts->eax;
				sc->sc_signode = n;
				ts->eip = (unsigned)act->sa_handler;
				ts->eflags &= ~(EFL_DF | EFL_TF);

				/* Mask out this signal and other signals as appropriate.
				   n->act->sa_mask always includes n->sigmask
				   (enforced by sigaction()).  */
				sig_mask |= act->sa_mask;
			}
		}
	}
}

/* Called when we took a trap from user mode
   and need to turn it into an appropriate signal.
   When (if) this routine returns,
   execution will be resumed after checking for newly posted signals.  */
void sig_trap(int dummy)
{
	struct trap_state *ts = (struct trap_state*)&dummy;
	int signo;

	switch (ts->trapno)
	{
		case 0 : signo = SIGFPE; break; /* divide by zero */
		case 1 : signo = SIGTRAP; break; /* debug exception */
		case 3 : signo = SIGTRAP; break; /* breakpoint */
		case 4 : signo = 16; break; /* into instruction (overflow) */
		case 5 : signo = 16; break; /* bound instruction */
		case 6 : signo = SIGILL; break; /* Invalid opcode */
		case 7 : signo = SIGFPE; break; /* coprocessor not available */
		case 8 : signo = 7; break; /* double fault */
		case 9 : signo = SIGSEGV; break; /* coprocessor segment overrun */
		case 10 : signo = SIGSEGV; break; /* Invalid TSS */
		case 11 : signo = SIGSEGV; break; /* Segment not present */
		case 12 : signo = SIGSEGV; break; /* stack exception */
		case 13 : signo = SIGSEGV; break; /* general protection */
		case 14 : signo = SIGSEGV; break; /* page fault */
		case 16 : signo = 7; break; /* coprocessor error */
		default: signo = 7;         /* "software generated"*/
	}

	kill(1, signo);
}

static void sig_gen_post_handler(struct signode *n)
{
	assert(n->signo > 0);
	assert(n->signo <= NSIG);
	assert(n == &sig_gen_nodes[n->signo - 1]);

	/* Mark the node available again.  */
	/* XXX If another signal of the same type comes
	   while the first is being handled, we'll miss the second.  */
	n->signo = 0;
}

int kill(int pid, int sig)
{
	if ((pid > 1) || (pid < 0))
		return ESRCH;
	if ((sig < 1) || (sig > NSIG))
		return EINVAL;

	/* If the signode is already pending, there's nothing to do.  */
	if (sig_gen_nodes[sig-1].signo)
		return 0;

	/* Initialize and queue the signode.  */
	sig_gen_nodes[sig-1].signo = sig;
	sig_gen_nodes[sig-1].sigmask = 1 << (sig - 1);
	sig_gen_nodes[sig-1].post_handler = sig_gen_post_handler;
	sig_insert_after(&sig_gen_nodes[sig-1], &sig_pending_list);

	new_sigs = 1;

	return 0;
}

int sigaction(int sig, struct sigaction *act, struct sigaction *oact)
{
	if ((sig <= 0) || (sig >= NSIG))
		return EINVAL;
	if (oact)
	{
		*oact = sigact[sig];
	}
	if (act)
	{
		sigact[sig] = *act;
		sigact[sig].sa_mask |= 1 << (sig - 1);
	}
	return 0;
}

int sigprocmask(int how, const sigset_t *set, sigset_t *oset)
{
	if (oset)
		*oset = sig_mask;
	if (set)
	{
		sigset_t nset;

		switch (how)
		{
			case SIG_BLOCK:
				nset = sig_mask | *set;
				break;
			case SIG_UNBLOCK:
				nset = sig_mask & ~*set;
				break;
			case SIG_SETMASK:
				nset = *set;
				break;
			default:
				return EINVAL;
		}
		if (nset & ~sig_mask)
			new_sigs = 1;
		sig_mask = nset;
	}
	return 0;
}

int sigpending(sigset_t *set)
{
	sigset_t s;
	struct signode *n;

	for (n = sig_pending_list.next; n != &sig_pending_list; n = n->next)
		s |= n->sigmask;

	*set = s;
	return 0;
}

int irq_request(int irq, int pri, struct sigevent *ev)
{
	if ((irq < 0) || (irq >= 16))
		return EINVAL;

	if (sig_irq_nodes[irq].signo)
		return EBUSY;

	if (ev->sigev_notify == SIGEV_SIGNAL)
	{
		int err;

		if ((ev->sigev_signo <= 0) || (ev->sigev_signo >= NSIG))
			return EINVAL;

		sig_irq_nodes[irq].signo = ev->sigev_signo;
		sig_irq_nodes[irq].sigmask = 1 << (ev->sigev_signo - 1);
		sig_irq_nodes[irq].value = ev->sigev_value;

		/* Interrupts can occur during this operation
		   (e.g. while fiddling with DOS interrupt vectors),
		   so the sig_irq_node must be ready for action by this time.  */
		if (err = progvec->irq_catch(irq, &sig_irq_nodes[irq]))
		{
			sig_irq_nodes[irq].signo = 0;
			return err;
		}
	}
	else if (ev->sigev_notify == SIGEV_FASTINT)
	{
		int err;

		/* Use sig_irq_nodes[irq].signo as a busy flag,
		   even though it is otherwise unused in this case.  */
		sig_irq_nodes[irq].signo = -1;

		if (err = progvec->irq_catch_fast(irq,
			(vm_offset_t)ev->sigev_handler,
			(vm_offset_t)ev->sigev_stack))
		{
			sig_irq_nodes[irq].signo = 0;
			return err;
		}
	}
	else
		return EINVAL;

	/* Enable the actual hardware interrupt,
	   saving the original state in the process.  */
	irq_orig_enable_mask &= ~(1 << irq);
	if (irq < 8)
	{
		unsigned char master_mask = inb(0x21);
		if (master_mask & (1 << irq))
			outb(0x21, master_mask & ~(1 << irq));
		else
			irq_orig_enable_mask |= 1 << irq;
	}
	else
	{
		unsigned char slave_mask = inb(0xa1);
		if (slave_mask & (1 << (irq-8)))
			outb(0xa1, slave_mask & ~(1 << (irq-8)));
		else
			irq_orig_enable_mask |= 1 << irq;
	}

	return 0;
}

int irq_free(int irq)
{
	if ((irq < 0) || (irq >= 16) || !sig_irq_nodes[irq].signo)
		return EINVAL;

	/* Disable the hardware IRQ
	   if it was disabled when the IRQ handler was installed.  */
	if (!(irq_orig_enable_mask & (1 << irq)))
	{
		if (irq < 8)
			outb(0x21, inb(0x21) | (1 << irq));
		else
			outb(0xa1, inb(0xa1) | (1 << (irq-8)));
	}

	if (sig_irq_nodes[irq].signo > 0)
	{
		/* Interrupts can occur during this operation
		   (e.g. while fiddling with DOS interrupt vectors),
		   so we must do this before clearing and disabling
		   the sig_irq_node.  */
		progvec->irq_free(irq);

		sig_clear(&sig_irq_nodes[irq]);
	}
	else
	{
		assert(sig_irq_nodes[irq].signo == -1);

		progvec->irq_free_fast(irq);
	}

	/* Mark the IRQ unused.  */
	sig_irq_nodes[irq].signo = 0;

	return 0;
}

void sig_cleanup(void)
{
	int i;

	for (i = 0; i < 16; i++)
		irq_free(i);
}

