/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996-1994 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>

#include "debug.h"
#include "prog.h"
#include "misc.h"
#include "dos_io.h"
#include "alloc.h"
#include "signal.h"
#include "vm_param.h"

static int sys__exit(void *args)
{
	struct {
		int rc;
	} *a = args;

	exit(a->rc);
}

static int sys_open(void *args)
{
	struct {
		const char *name;
		int flags;
		int mode;
	} *a = args;

	return open(progva_to_va(a->name), a->flags, a->mode);
}

static int sys_close(void *args)
{
	struct {
		unsigned fd;
	} *a = args;

	return close(a->fd);
}

static int sys_read(void *args)
{
	struct {
		int fd;
		char *buf;
		int size;
	} *a = args;

	return read(a->fd, progva_to_va(a->buf), a->size);
}

static int sys_write(void *args)
{
	struct {
		int fd;
		const char *buf;
		int size;
	} *a = args;

	return write(a->fd, progva_to_va(a->buf), a->size);
}

static int sys_lseek(void *args)
{
	struct {
		int fd;
		int offset;
		int whence;
	} *a = args;

	return lseek(a->fd, a->offset, a->whence);
}

static int sys_rename(void *args)
{
	struct {
		const char *oldpath;
		const char *newpath;
	} *a = args;

	errno = dos_rename(progva_to_va(a->oldpath), progva_to_va(a->newpath));
	return errno ? -1 : 0;
}

static int sys_sbrk(void *args)
{
	struct {
		int mem_needed;
	} *a = args;
	vm_offset_t prog_mem_va;

	errno = progvec->sbrk(a->mem_needed, &prog_mem_va);
	return errno ? -1 : prog_mem_va;
}

static int sys_gettimeofday(void *args)
{
	struct {
		struct timeval *tv;
		struct timezone *tz;
	} *a = args;

	errno = dos_gettimeofday(a->tv ? progva_to_va(a->tv) : 0,
			         a->tz ? progva_to_va(a->tz) : 0);
	return errno ? -1 : 0;
}

static int sys_unlink(void *args)
{
	struct {
		const char *filename;
	} *a = args;

	errno = dos_unlink(progva_to_va(a->filename));
	return errno ? -1 : 0;
}

static int sys_fstat(void *args)
{
	struct {
		int fd;
		struct stat *buf;
	} *a = args;

	return fstat(a->fd, progva_to_va(a->buf));
}

static int sys_mlock(void *args)
{
	struct {
		vm_offset_t addr;
		vm_size_t len;
	} *a = args;

	errno = progvec->mlock(a->addr, a->len);
	return errno ? -1 : 0;
}

static int sys_munlock(void *args)
{
	struct {
		vm_offset_t addr;
		vm_size_t len;
	} *a = args;

	errno = progvec->munlock(a->addr, a->len);
	return errno ? -1 : 0;
}

static int sys_mmap(void *args)
{
	struct {
		void *addr;
		size_t len;
		int prot;
		int flags;
		int fd;
		off_t off;
	} *a = args;
	caddr_t addr;

	addr = mmap(a->addr, a->len, a->prot, a->flags, a->fd, a->off);
	return addr;
}

static int sys_munmap(void *args)
{
	return 0;
}

static int sys_irq_request(void *args)
{
	struct {
		int irq;
		int pri;
		struct sigevent *ev;
	} *a = args;

	errno = irq_request(a->irq, a->pri, progva_to_va(a->ev));
	return errno ? -1 : 0;
}

static int sys_irq_free(void *args)
{
	struct {
		int irq;
	} *a = args;

	errno = irq_free(a->irq);
	return errno ? -1 : 0;
}

static int sys_kill(void *args)
{
	struct {
		int pid;
		int sig;
	} *a = args;

	errno = kill(a->pid, a->sig);
	return errno ? -1 : 0;
}

static int sys_sigaction(void *args)
{
	struct {
		int sig;
		struct sigaction *act;
		struct sigaction *oact;
	} *a = args;

	errno = sigaction(a->sig,
			  a->act ? progva_to_va(a->act) : 0,
			  a->oact ? progva_to_va(a->oact) : 0);
	return errno ? -1 : 0;
}

static int sys_sigprocmask(void *args)
{
	struct {
		int how;
		const sigset_t *set;
		sigset_t *oset;
	} *a = args;

	errno = sigprocmask(a->how,
			    a->set ? progva_to_va(a->set) : 0,
			    a->oset ? progva_to_va(a->oset) : 0);
	return errno ? -1 : 0;
}

static int sys_sigpending(void *args)
{
	struct {
		const sigset_t *sigmask;
	} *a = args;

	errno = sigpending(progva_to_va(a->sigmask));
	return errno ? -1 : 0;
}

static int sys_alloc_dos(void *args)
{
	struct {
		vm_size_t size;
		int align_bits;
		vm_offset_t align_ofs;
	} *a = args;
	void *buf;

	buf = alloc_dos(a->size, a->align_bits,
			phystokv(a->align_ofs) & ~((1 << a->align_bits)-1));
	if (!buf) return 0;
#if 0
	printf("alloc_dos(%d,%d,%d) -> phys %08x kv %08x pv %08x\n",
		a->size, a->align_bits, a->align_ofs,
		kvtophys(buf), buf, va_to_progva(buf));
#endif
	return (int)kvtophys(buf);
}

static int sys_alloc_dma(void *args)
{
	struct {
		vm_size_t size;
		int align_bits;
		vm_offset_t align_ofs;
	} *a = args;
	void *buf;

	buf = alloc_dma(a->size, a->align_bits,
			phystokv(a->align_ofs) & ~((1 << a->align_bits)-1));
	if (!buf) return 0;
	return (int)kvtophys(buf);
}

static int sys_mlockall(void *args)
{
	struct {
		int flags;
	} *a = args;

	errno = progvec->mlockall(a->flags);
	return errno ? -1 : 0;
}

static int sys_munlockall(void *args)
{
	errno = progvec->munlockall();
	return errno ? -1 : 0;
}

int (*syscall_table[])(void *args) = {
#define s(name) sys_##name,
#include <moss/syscall_list.h>
};

