/*
 * MOSS: DOS extender based on the Flux OS toolkit
 * Copyright (C) 1996 Sleepless Software
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	Author: Bryan Ford
 */

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/stat.h>

#include "misc.h"
#include "dos_io.h"
#include "efs.h"
#include "prog.h"

struct fd_file
{
	dos_fd_t dos_fd;
};

struct fd
{
	int type;
	union
	{
		struct fd_file f;
		struct fd_efs e;
	} u;
};
#define FD_UNUSED	0
#define FD_FILE		1
#define FD_EFS		2
#define FD_DEV_MEM	3

static unsigned nfds = 20; /*XXX*/
static struct fd fds[20];

void vfs_init()
{
	fds[0].type = FD_FILE; /* XXX tty? */
	fds[0].u.f.dos_fd = 0;
	fds[1].type = FD_FILE; /* XXX tty? */
	fds[1].u.f.dos_fd = 1;
	fds[2].type = FD_FILE; /* XXX tty? */
	fds[2].u.f.dos_fd = 2;
}

static int getfd(int *out_fd)
{
	unsigned i;

	for (i = 0; i < nfds; i++)
	{
		if (fds[i].type == FD_UNUSED)
		{
			*out_fd = i;
			return 0;
		}
	}
	return EMFILE;
}

#define CHECKFD(fd) \
	({ if (((fd) >= nfds) || (fds[fd].type == FD_UNUSED))	\
		{ errno = EBADF; return -1; } })

int open(const char *name, int flags, ...)
{
	mode_t mode;
	int fd;

	/* Find the mode parameter.  */
	va_list vl;
	va_start(vl, flags); 
	mode = va_arg(vl, mode_t);
	va_end(vl);

	/* Allocate an unused file descriptor.  */
	if ((errno = getfd(&fd)) != 0)
		return -1;

	/* Horrible kludge: support /dev/mem for mmap.  */
	if (!strcmp(name, "/dev/mem"))
	{
		fds[fd].type = FD_DEV_MEM;
		return fd;
	}

	/* Try opening the file as a DOS file.  */
	/* XXX do slash translation */
	/* XXX tty handling */
	errno = dos_open(name, flags, mode, &fds[fd].u.f.dos_fd);
	if (errno == 0) /* != ENOENT */
	{
		fds[fd].type = FD_FILE;
		return fd;
	}

	/* See if we have an embedded file of that name.  */
	if (efs_open(name, flags, mode, &fds[fd].u.e) == 0)
	{
		fds[fd].type = FD_EFS;
		return fd;
	}

	return -1;
}

int close(int fd)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_FILE:
		{
			if ((errno = dos_close(fds[fd].u.f.dos_fd)) != 0)
				return -1;
			break;
		}
		case FD_EFS:
		{
			if ((errno = efs_close(&fds[fd].u.e)) != 0)
				return -1;
			break;
		}
	}
	fds[fd].type = FD_UNUSED;
	return 0;
}

int read(int fd, void *buf, size_t size)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_FILE:
		{
			vm_size_t actual;

			errno = dos_read(fds[fd].u.f.dos_fd, buf, size,
					 &actual);
			return errno ? -1 : actual;
		}
		case FD_EFS:
		{
			size_t actual;

			errno = efs_read(&fds[fd].u.e, buf, size, &actual);
			return errno ? -1 : actual;
		}
	}
	errno = EINVAL;
	return -1;
}

int write(int fd, const void *buf, size_t size)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_FILE:
		{
			vm_size_t actual;

			errno = dos_write(fds[fd].u.f.dos_fd, buf, size,
					  &actual);
			return errno ? -1 : actual;
		}
	}
	errno = EINVAL;
	return -1;
}

off_t lseek(int fd, off_t offset, int whence)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_FILE:
		{
			vm_offset_t newpos;

			errno = dos_seek(fds[fd].u.f.dos_fd,
					  offset, whence, &newpos);
			return errno ? -1 : newpos;
		}
		case FD_EFS:
		{
			off_t newpos;

			errno = efs_seek(&fds[fd].u.e, offset, whence, &newpos);
			return errno ? -1 : newpos;
		}
	}
	errno = EINVAL;
	return -1;
}

int fstat(int fd, struct stat *buf)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_FILE:
		{
			errno = dos_fstat(fds[fd].u.f.dos_fd, buf);
			return errno ? -1 : 0;
		}
		case FD_EFS:
		{
			errno = efs_fstat(&fds[fd].u.e, buf);
			return errno ? -1 : 0;
		}
	}
	errno = EINVAL;
	return -1;
}

caddr_t mmap(caddr_t pva, size_t len, int prot, int flags,
	     int fd, off_t offset)
{
	CHECKFD(fd);
	switch (fds[fd].type)
	{
		case FD_DEV_MEM:
		{
			/*if (flags & (MAP_FIXED | MAP_PRIVATE))
				return EINVAL;*/
			errno = progvec->mmap_phys(offset, len, &pva);
			return errno ? (caddr_t)-1 : pva;
		}
	}
	errno = EINVAL;
	return (caddr_t)-1;
}

